﻿using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using ProcessHacker.Native.Security.AccessControl;
using ProcessHacker.Native;
using ProcessHacker.Native.Api;
using ProcessHacker.Native.Security;
using System.Security.Cryptography.X509Certificates;
using System.Net;
using System.Net.Sockets;


namespace HttpSysManager.HttpAPI
{
	public abstract class HttpAPIQuery<TOutput>
	{

		public Type OutputType
		{
			get
			{
				return typeof(TOutput);
			}
		}

		public IEnumerable<TOutput> Execute()
		{
			using(Functions.InitializeConfig())
			{
				UnManagedOutputStruct<TOutput> result;
				uint dwToken = 0;
				while(ExecuteCore(out result, dwToken))
				{
					using(result)
					{
						yield return result.Value;
						dwToken++;
					}
				}
			}
		}

		protected abstract bool ExecuteCore(out UnManagedOutputStruct<TOutput> result, uint dwToken);
	}

	public class UrlACLQuery : HttpAPIQuery<HTTP_SERVICE_CONFIG_URLACL_SET>
	{
		public string Url
		{
			get;
			set;
		}

		protected override bool ExecuteCore(out UnManagedOutputStruct<HTTP_SERVICE_CONFIG_URLACL_SET> result, uint dwToken)
		{
			result = new UnManagedOutputStruct<HTTP_SERVICE_CONFIG_URLACL_SET>();
			if(Url != null && dwToken != 0)
				return false;
			return Functions.HttpQueryACLServiceConfiguration(IntPtr.Zero,
				new HTTP_SERVICE_CONFIG_URLACL_QUERY()
				{
					QueryDesc = Url == null ? HTTP_SERVICE_CONFIG_QUERY_TYPE.HttpServiceConfigQueryNext : HTTP_SERVICE_CONFIG_QUERY_TYPE.HttpServiceConfigQueryExact,
					KeyDesc = new HTTP_SERVICE_CONFIG_URLACL_KEY(Url),
					dwToken = dwToken,
				},
				out result);
		}
	}

	public class SSLInfoQuery : HttpAPIQuery<HTTP_SERVICE_CONFIG_SSL_SET>
	{
		protected override bool ExecuteCore(out UnManagedOutputStruct<HTTP_SERVICE_CONFIG_SSL_SET> result, uint dwToken)
		{
			result = new UnManagedOutputStruct<HTTP_SERVICE_CONFIG_SSL_SET>();
			if(Endpoint != null && dwToken != 0)
				return false;
			return Functions.HttpQuerySSLServiceConfiguration(IntPtr.Zero,
				new HTTP_SERVICE_CONFIG_SSL_QUERY()
				{
					QueryDesc = this.Endpoint == null ? HTTP_SERVICE_CONFIG_QUERY_TYPE.HttpServiceConfigQueryNext : HTTP_SERVICE_CONFIG_QUERY_TYPE.HttpServiceConfigQueryExact,
					KeyDesc = new HTTP_SERVICE_CONFIG_SSL_KEY(Endpoint),
					dwToken = dwToken,
				},
				out result);
		}

		public IPEndPoint Endpoint
		{
			get;
			set;
		}
	}

	public class UrlAcl : ISecurable
	{
		HTTP_SERVICE_CONFIG_URLACL_SET _Acl;
		public UrlAcl(HTTP_SERVICE_CONFIG_URLACL_SET acl)
		{
			Update(acl);
		}

		private void Update(HTTP_SERVICE_CONFIG_URLACL_SET acl)
		{
			_Acl = acl;
			SecurityDescriptor = SecurityDescriptor.FromSDDL(acl.ParamDesc.pStringSecurityDescriptor);
		}
		public string Prefix
		{
			get
			{
				return _Acl.KeyDesc.pUrlPrefix;
			}
		}
		public SecurityDescriptor SecurityDescriptor
		{
			get;
			private set;
		}
		public override string ToString()
		{
			return _Acl.ToString();
		}

		#region ISecurable Members

		SecurityDescriptor ISecurable.GetSecurity(ProcessHacker.Native.Api.SecurityInformation securityInformation)
		{
			return SecurityDescriptor;
		}

		public void SetSecurity(SecurityDescriptor securityDescriptor)
		{
			((ISecurable)this).SetSecurity(SecurityInformation.Dacl, securityDescriptor);
		}
		void ISecurable.SetSecurity(ProcessHacker.Native.Api.SecurityInformation securityInformation, SecurityDescriptor securityDescriptor)
		{
			if(securityDescriptor.ToString() == SecurityDescriptor.Empty.ToString())
				return;
			var manager = new HttpAPIManager();
			manager.SetUrlAcl(Prefix, securityDescriptor);
			Update(manager.GetAclInfo(Prefix)._Acl);
		}



		#endregion
	}

	public class IntPtrAsArray
	{
		public IntPtrAsArray(IntPtr ptr)
		{
			_Ptr = ptr;
		}
		private readonly IntPtr _Ptr;
		public IntPtr Ptr
		{
			get
			{
				return _Ptr;
			}
		}
		public byte this[int index]
		{
			get
			{
				return Marshal.ReadByte(Ptr.Increment(index));
			}
			set
			{
				Marshal.WriteByte(Ptr.Increment(index), value);
			}
		}
	}
	public class SSLInfo
	{
		public SSLInfo()
			: this(StoreName.My, null)
		{
			StoreName = StoreName.My;
		}
		public SSLInfo(string thumbprint)
			: this(StoreName.My, thumbprint)
		{

		}
		public SSLInfo(StoreName storeName, string thumbprint)
		{
			StoreName = storeName;
			if(thumbprint != null)
				SetCertificate(storeName, thumbprint);
		}
		public IPEndPoint Endpoint
		{
			get;
			set;
		}

		public void SetCertificate(StoreName store, string thumbprint)
		{
			if(thumbprint == null)
				throw new ArgumentNullException("thumbprint");
			Certificate = CertificateHelper.GetCertificate(StoreLocation.LocalMachine, store, X509FindType.FindByThumbprint, thumbprint);
			if(Certificate == null)
				throw new ArgumentException("Certificate " + thumbprint + " not found in the store " + Enum.GetName(typeof(StoreName), store) + " of the local machine");
		}
		public X509Certificate2 Certificate
		{
			get;
			private set;
		}
		public Guid AppId
		{
			get;
			set;
		}
		public StoreName StoreName
		{
			get;
			set;
		}

		public bool NegotiateClientCert
		{
			get;
			set;
		}
		public bool CertificateAccountMapping
		{
			get;
			set;
		}



		public static SSLInfo Create(HTTP_SERVICE_CONFIG_SSL_SET set)
		{

			var param = set.ParamDesc;
			SSLInfo info = new SSLInfo();
			info._InitialConfig = set;
			info.AppId = param.AppId;
			info.Endpoint = Create(new IntPtrAsArray(set.KeyDesc.pIpPort));
			if(string.IsNullOrEmpty(param.pSslCertStoreName))
				info.StoreName = StoreName.My;
			else
				info.StoreName = (StoreName)Enum.Parse(typeof(StoreName), param.pSslCertStoreName, true);
			info.NegotiateClientCert = (param.DefaultFlags & Constants.HTTP_SERVICE_CONFIG_SSL_FLAG_NEGOTIATE_CLIENT_CERT) != 0;
			info.CertificateAccountMapping = (param.DefaultFlags & Constants.HTTP_SERVICE_CONFIG_SSL_FLAG_USE_DS_MAPPER) != 0;
			var store = new X509Store(info.StoreName, StoreLocation.LocalMachine);
			store.Open(OpenFlags.ReadOnly);
			try
			{
				info.Certificate = store.Certificates.Find(X509FindType.FindByThumbprint, GetHashstring(param), false).OfType<X509Certificate2>().FirstOrDefault();
			}
			finally
			{
				store.Close();
			}
			return info;
		}
		internal uint SetCurrentConfig()
		{
			IntPtr strPtr = IntPtr.Zero;
			try
			{
				_InitialConfig.KeyDesc = new HTTP_SERVICE_CONFIG_SSL_KEY(Endpoint);
				_InitialConfig.ParamDesc.AppId = AppId;
				_InitialConfig.ParamDesc.pSslCertStoreName = Enum.GetName(typeof(StoreName), StoreName).ToUpper();
				if(NegotiateClientCert)
					_InitialConfig.ParamDesc.DefaultFlags |= Constants.HTTP_SERVICE_CONFIG_SSL_FLAG_NEGOTIATE_CLIENT_CERT;
				else
					_InitialConfig.ParamDesc.DefaultFlags &= ~Constants.HTTP_SERVICE_CONFIG_SSL_FLAG_NEGOTIATE_CLIENT_CERT;

				if(CertificateAccountMapping)
					_InitialConfig.ParamDesc.DefaultFlags |= Constants.HTTP_SERVICE_CONFIG_SSL_FLAG_USE_DS_MAPPER;
				else
					_InitialConfig.ParamDesc.DefaultFlags &= ~Constants.HTTP_SERVICE_CONFIG_SSL_FLAG_USE_DS_MAPPER;
				strPtr = SetHashstring(ref _InitialConfig.ParamDesc, Certificate.Thumbprint);
				var result = HttpAPI.Functions.HttpSetSSLServiceConfiguration(IntPtr.Zero, HTTP_SERVICE_CONFIG_ID.HttpServiceConfigSSLCertInfo, ref _InitialConfig, Marshal.SizeOf(_InitialConfig), IntPtr.Zero);
				return result;
			}
			finally
			{
				if(_InitialConfig.KeyDesc.pIpPort != IntPtr.Zero)
				{
					Marshal.FreeHGlobal(_InitialConfig.KeyDesc.pIpPort);
					_InitialConfig.KeyDesc.pIpPort = IntPtr.Zero;
				}
				if(strPtr != IntPtr.Zero)
					Marshal.FreeHGlobal(strPtr);
			}
		}


		static IPEndPoint Create(IntPtrAsArray socketAddress)
		{

			var family = (AddressFamily)((int)socketAddress[0] | (int)socketAddress[1] << 8);
			if(family == AddressFamily.InterNetworkV6)
			{
				byte[] array = new byte[16];
				for(int i = 0 ; i < array.Length ; i++)
				{
					array[i] = socketAddress[i + 8];
				}
				int port = ((int)socketAddress[2] << 8 & 65280) | (int)socketAddress[3];
				long scopeid = (long)(((int)socketAddress[27] << 24) + ((int)socketAddress[26] << 16) + ((int)socketAddress[25] << 8) + (int)socketAddress[24]);
				return new IPEndPoint(new IPAddress(array, scopeid), port);
			}
			if(family != AddressFamily.InterNetwork)
				throw new NotSupportedException(family.ToString());
			int port2 = ((int)socketAddress[2] << 8 & 65280) | (int)socketAddress[3];
			long address = (long)((int)(socketAddress[4] & 255) | ((int)socketAddress[5] << 8 & 65280) | ((int)socketAddress[6] << 16 & 16711680) | (int)socketAddress[7] << 24) & unchecked((long)unchecked((ulong)-1));
			return new IPEndPoint(address, port2);
		}

		private static string GetHashstring(HTTP_SERVICE_CONFIG_SSL_PARAM param)
		{
			byte[] data = new byte[param.SslHashLength];
			for(int i = 0 ; i < param.SslHashLength ; i++)
			{
				var currentPtr = IntPtrExtensions.Increment(param.pSslHash, i);
				data[i] = Marshal.ReadByte(currentPtr);
			}
			return BitConverter.ToString(data).Replace("-", " ");
		}

		private IntPtr SetHashstring(ref HTTP_SERVICE_CONFIG_SSL_PARAM param, string value)
		{
			var array = Enumerable.Range(0, value.Length)
					 .Where(x => x % 2 == 0)
					 .Select(x => Convert.ToByte(value.Substring(x, 2), 16))
					 .ToArray();
			param.SslHashLength = array.Length;
			var arrayPtr = Marshal.AllocHGlobal(array.Length);
			Marshal.Copy(array, 0, arrayPtr, array.Length);
			param.pSslHash = arrayPtr;
			return arrayPtr;
		}

		private HTTP_SERVICE_CONFIG_SSL_SET _InitialConfig;


	}



	public class HttpAPIManager
	{
		public IEnumerable<UrlAcl> GetAclInfo()
		{
			return new UrlACLQuery().Execute().Select(a => new UrlAcl(a)).ToList();
		}

		public void SetUrlAcl(string url, SecurityDescriptor securityDescriptor)
		{
			using(new HttpInitialize(Constants.HTTPAPI_VERSION_2, Constants.HTTP_INITIALIZE_CONFIG))
			{
				if(securityDescriptor == null)
				{
					RemoveUrlAcl(url);
					return;
				}
				var acl = new HTTP_SERVICE_CONFIG_URLACL_SET(url, securityDescriptor.ToString());
				var result = Functions.HttpSetACLServiceConfiguration(IntPtr.Zero, HTTP_SERVICE_CONFIG_ID.HttpServiceConfigUrlAclInfo, ref acl, Marshal.SizeOf(acl), IntPtr.Zero);
				if(result != 0 && result != (int)ErrorCodes.ERROR_ALREADY_EXISTS)
					Win32.Throw((int)result);
				if(result == (int)ErrorCodes.ERROR_ALREADY_EXISTS)
				{
					RemoveUrlAcl(url);
					SetUrlAcl(url, securityDescriptor);
				}
			}
		}

		public void RemoveUrlAcl(string url)
		{
			using(new HttpInitialize(Constants.HTTPAPI_VERSION_2, Constants.HTTP_INITIALIZE_CONFIG))
			{
				var acl = new HTTP_SERVICE_CONFIG_URLACL_SET(url, SecurityDescriptor.Empty.ToString());
				var result = Functions.HttpDeleteACLServiceConfiguration(IntPtr.Zero, HTTP_SERVICE_CONFIG_ID.HttpServiceConfigUrlAclInfo, ref acl, Marshal.SizeOf(acl), IntPtr.Zero);
				if(result != 0)
					Win32.Throw((int)result);
			}
		}
		static protected void CreateBuffer(object initObject, out IntPtr ptr, out int lenght)
		{
			lenght = Marshal.SizeOf(initObject);
			ptr = Marshal.AllocHGlobal(lenght);
			Marshal.StructureToPtr(initObject, ptr, false);
		}


		public UrlAcl GetAclInfo(string url)
		{
			return new UrlACLQuery()
			{
				Url = url
			}.Execute().Select(a => new UrlAcl(a)).ToList().FirstOrDefault();
		}

		public IEnumerable<SSLInfo> GetSSLInfos()
		{
			return new SSLInfoQuery().Execute().Select(a => SSLInfo.Create(a)).ToList();
		}

		public void SetSSLInfo(IPEndPoint endpoint, SSLInfo info)
		{
			using(new HttpInitialize(Constants.HTTPAPI_VERSION_2, Constants.HTTP_INITIALIZE_CONFIG))
			{
				if(info == null)
				{
					RemoveSSLInfo(endpoint);
				}
				else
				{
					info.Endpoint = endpoint;
					uint result = info.SetCurrentConfig();
					if(result == (int)ErrorCodes.ERROR_NO_SUCH_LOGON_SESSION)
						throw new ArgumentException("Invalid Certificate");
					if(result != 0 && result != (int)ErrorCodes.ERROR_ALREADY_EXISTS)
						Win32.Throw((int)result);
					if(result == (int)ErrorCodes.ERROR_ALREADY_EXISTS)
					{
						RemoveSSLInfo(endpoint);
						SetSSLInfo(endpoint, info);
					}
				}
			}
		}

		public void RemoveSSLInfo(IPEndPoint endpoint)
		{
			using(new HttpInitialize(Constants.HTTPAPI_VERSION_2, Constants.HTTP_INITIALIZE_CONFIG))
			{
				var sslInfo = new HTTP_SERVICE_CONFIG_SSL_SET(endpoint);
				try
				{
					var result = Functions.HttpDeleteSSLServiceConfiguration(IntPtr.Zero, HTTP_SERVICE_CONFIG_ID.HttpServiceConfigSSLCertInfo, ref sslInfo, Marshal.SizeOf(sslInfo), IntPtr.Zero);
					if(result != 0)
						Win32.Throw((int)result);
				}
				finally
				{
					if(sslInfo.KeyDesc.pIpPort != IntPtr.Zero)
						Marshal.FreeHGlobal(sslInfo.KeyDesc.pIpPort);
				}
			}
		}

		public SSLInfo GetSSLInfo(IPEndPoint endpoint)
		{
			return new SSLInfoQuery()
			{
				Endpoint = endpoint
			}.Execute().Select(a => SSLInfo.Create(a)).ToList().FirstOrDefault();
		}
	}
}
